#include "config.h"
#include <Arduino.h>
#include <EEPROM.h>
#include <SoftwareSerial.h>
#include <ESP8266WiFi.h>
#include <LittleFS.h>
#include <ArduinoOTA.h>
#include <ESP8266mDNS.h>
#include <ESP8266WebServer.h>
#include <WebSocketsServer.h>
#include <Wire.h>
#include <PMserial.h>
#include <bsec.h>
#include <PubSubClient.h>
#include <SimpleMap.h>
#include <ArduinoJson.h>

#define PMS_TX D3
#define PMS_RX D4
#define STATE_SAVE_PERIOD	UINT32_C(360 * 60 * 1000) // 360 minutes - 4 times a day

SoftwareSerial swSerial;
WiFiClient wifiClient;
MDNSResponder mdns;
ESP8266WebServer server(80);
WebSocketsServer webSocket = WebSocketsServer(81);
SerialPM pms(PMS5003, PMS_RX, PMS_TX);
Bsec iaqSensor;
PubSubClient mqttClient(wifiClient);
IPAddress mqttServer;
SimpleMap<String, double> *sensorData;

const uint8_t bsec_config_iaq[] = {
#include "config/generic_33v_3s_4d/bsec_iaq.txt"
};
uint8_t bsecState[BSEC_MAX_STATE_BLOB_SIZE] = {0};
uint16_t stateUpdateCounter = 0;

void logLine(String line, bool newline = true)
{
  Serial.print(line);
  if (newline)
  {
    Serial.println();
  }
}

void setupWifi()
{
  logLine("Connecting to '" + String(WIFI_SSID) + "' ", false);
  WiFi.hostname(HOSTNAME);
  WiFi.begin(WIFI_SSID, WIFI_PSK);

  int i = 50;
  while (i > 0)
  { // check for roughly 10 seconds (i = 50 * 200ms)
    Serial.print(".");
    if (WiFi.status() != WL_CONNECTED)
    {
      i--;
      delay(200);
    }
    else
    {
      i = 0;
    }
  }
  if (WiFi.status() == WL_CONNECTED)
  {
    Serial.println();
    logLine("WiFi connected with IP: " + WiFi.localIP().toString());
  }
  else
  {
    logLine("WiFi connection failed.");
  }
}

void setupMDNS() {
  if (!MDNS.begin(HOSTNAME)) {
    Serial.println("Error setting up MDNS responder!");
  }
  MDNS.addService("http", "tcp", 80);
}

String getContentType(String filename)
{ // convert the file extension to the MIME type
  if (filename.endsWith(".html"))
    return "text/html";
  else if (filename.endsWith(".css"))
    return "text/css";
  else if (filename.endsWith(".js"))
    return "application/javascript";
  else if (filename.endsWith(".ico"))
    return "image/x-icon";
  else if (filename.endsWith(".gz"))
    return "application/x-gzip";
  return "text/plain";
}

bool handleFileRead(String path)
{ // send the right file to the client (if it exists)
  Serial.println("handleFileRead: " + path);
  if (path.endsWith("/"))
    path += "index.html";                    // If a folder is requested, send the index file
  String contentType = getContentType(path); // Get the MIME type
  String pathWithGz = path + ".gz";
  if (LittleFS.exists(pathWithGz) || LittleFS.exists(path))
  {                                                     // If the file exists, either as a compressed archive, or normal
    if (LittleFS.exists(pathWithGz))                    // If there's a compressed version available
      path += ".gz";                                    // Use the compressed version
    File file = LittleFS.open(path, "r");               // Open the file
    size_t sent = server.streamFile(file, contentType); // Send it to the client
    file.close();                                       // Close the file again
    Serial.printf("Sent file: %s (%d)\r\n", path.c_str(), sent);
    return true;
  }
  Serial.println(String("\tFile Not Found: ") + path);
  return false; // If the file doesn't exist, return false
}

void sendWebsocketData(uint8_t num)
{
  DynamicJsonDocument json(1024);
  for (int i = 0; i < sensorData->size(); i++)
  {
    json[sensorData->getKey(i)] = sensorData->getData(i);
  }
  const size_t strsize = measureJson(json) + 1;
  char MSG[strsize];
  serializeJson(json, MSG, strsize);
  webSocket.broadcastTXT(MSG, strsize - 1);
}

void webSocketEvent(uint8_t num, WStype_t type, uint8_t *payload, size_t length)
{
  Serial.printf("webSocketEvent(%d, %d, ...)\r\n", num, type);
  switch (type)
  {
  case WStype_DISCONNECTED:
    Serial.printf("[%u] Disconnected!\r\n", num);
    break;
  case WStype_CONNECTED:
  {
    IPAddress ip = webSocket.remoteIP(num);
    Serial.printf("[%u] Connected from %d.%d.%d.%d url: %s\r\n", num, ip[0], ip[1], ip[2], ip[3], payload);
    sendWebsocketData(num);
  }
  break;
  case WStype_TEXT:
    Serial.printf("[%u] get Text: %s\r\n", num, payload);
    break;
  case WStype_BIN:
    Serial.printf("[%u] get binary length: %u\r\n", num, length);
    hexdump(payload, length);
    break;
  case WStype_PING:
    Serial.printf("[%u] ping received\r\n", num);
    break;
  case WStype_PONG:
    Serial.printf("[%u] pong received\r\n", num);
    break;
  default:
    Serial.printf("Invalid WStype [%d]\r\n", type);
    break;
  }
}

void setupWebserver()
{
  LittleFS.begin();
  server.onNotFound([]() {                              // If the client requests any URI
    if (!handleFileRead(server.uri()))                  // send it if it exists
      server.send(404, "text/plain", "404: Not Found"); // otherwise, respond with a 404 (Not Found) error
  });
  server.begin();

  webSocket.begin();
  webSocket.onEvent(webSocketEvent);
}

void messageReceived(String &topic, String &payload)
{
  Serial.println("incoming: " + topic + " - " + payload);
}

void connectMQTT()
{
  while (!mqttClient.connected())
  {
    if (mqttClient.connect(HOSTNAME, MQTT_USER, MQTT_PASS))
    {
      Serial.println("MQTT connected");
    }
    else
    {
      Serial.print("MQTT connect failed, rc=");
      Serial.println(mqttClient.state());
      delay(200);
    }
  }
}

void setupMQTT()
{
  if (mqttServer.fromString(MQTT_HOST))
  {
    mqttClient.setServer(mqttServer, 1883);
    mqttClient.setBufferSize(512);
  }
  else
  {
    logLine("Failed to set IP for MQTT server");
  }
  connectMQTT();
}

void setupOTA()
{
  ArduinoOTA.setHostname(HOSTNAME);
  ArduinoOTA.onStart([]()
                     {
                       String type;
                       if (ArduinoOTA.getCommand() == U_FLASH)
                       {
                         type = "sketch";
                       }
                       else
                       { // U_SPIFFS
                         type = "filesystem";
                         LittleFS.end();
                       }
                       Serial.println("Start updating " + type);
                     });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total)
                        { Serial.printf("Progress: %u%%\r", (progress / (total / 100))); });
  ArduinoOTA.onEnd([]()
                   { Serial.println("\nEnd"); });
  ArduinoOTA.onError([](ota_error_t error)
                     {
                       Serial.printf("Error[%u]: ", error);
                       if (error == OTA_AUTH_ERROR)
                         Serial.println("Auth Failed");
                       else if (error == OTA_BEGIN_ERROR)
                         Serial.println("Begin Failed");
                       else if (error == OTA_CONNECT_ERROR)
                         Serial.println("Connect Failed");
                       else if (error == OTA_RECEIVE_ERROR)
                         Serial.println("Receive Failed");
                       else if (error == OTA_END_ERROR)
                         Serial.println("End Failed");
                     });
  ArduinoOTA.begin();
}

void setupPms()
{
  swSerial.begin(9600, SWSERIAL_8N1, PMS_RX, PMS_TX);
  pms.init();
}

void checkIaqSensorStatus(void)
{
  if (iaqSensor.status != BSEC_OK)
  {
    if (iaqSensor.status < BSEC_OK)
    {
      logLine("BSEC error code : " + String(iaqSensor.status));
    }
    else
    {
      logLine("BSEC warning code : " + String(iaqSensor.status));
    }
  }

  if (iaqSensor.bme680Status != BME680_OK)
  {
    if (iaqSensor.bme680Status < BME680_OK)
    {
      logLine("BME680 error code : " + String(iaqSensor.bme680Status));
    }
    else
    {
      logLine("BME680 warning code : " + String(iaqSensor.bme680Status));
    }
  }
}

void loadBmeState(void)
{
  if (EEPROM.read(0) == BSEC_MAX_STATE_BLOB_SIZE) {
    // Existing state in EEPROM
    Serial.println("Reading state from EEPROM");

    for (uint8_t i = 0; i < BSEC_MAX_STATE_BLOB_SIZE; i++) {
      bsecState[i] = EEPROM.read(i + 1);
      Serial.println(bsecState[i], HEX);
    }

    iaqSensor.setState(bsecState);
    checkIaqSensorStatus();
  } else {
    // Erase the EEPROM with zeroes
    Serial.println("Erasing EEPROM");

    for (uint8_t i = 0; i < BSEC_MAX_STATE_BLOB_SIZE + 1; i++)
      EEPROM.write(i, 0);

    EEPROM.commit();
  }
}

void updateBmeState(void)
{
  bool update = false;
  /* Set a trigger to save the state. Here, the state is saved every STATE_SAVE_PERIOD with the first state being saved once the algorithm achieves full calibration, i.e. iaqAccuracy = 3 */
  if (stateUpdateCounter == 0) {
    if (iaqSensor.iaqAccuracy >= 3) {
      update = true;
      stateUpdateCounter++;
    }
  } else {
    /* Update every STATE_SAVE_PERIOD milliseconds */
    if ((stateUpdateCounter * STATE_SAVE_PERIOD) < millis()) {
      update = true;
      stateUpdateCounter++;
    }
  }

  if (update) {
    iaqSensor.getState(bsecState);
    checkIaqSensorStatus();

    Serial.println("Writing state to EEPROM");

    for (uint8_t i = 0; i < BSEC_MAX_STATE_BLOB_SIZE ; i++) {
      EEPROM.write(i + 1, bsecState[i]);
      Serial.println(bsecState[i], HEX);
    }

    EEPROM.write(0, BSEC_MAX_STATE_BLOB_SIZE);
    EEPROM.commit();
  }
}

void setupBme()
{
  EEPROM.begin(BSEC_MAX_STATE_BLOB_SIZE + 1);
  Wire.begin();
  iaqSensor.begin(BME680_I2C_ADDR_PRIMARY, Wire);
  checkIaqSensorStatus();
  iaqSensor.setConfig(bsec_config_iaq);
  checkIaqSensorStatus();
  loadBmeState();
  bsec_virtual_sensor_t sensorList[10] = {
      BSEC_OUTPUT_RAW_TEMPERATURE,
      BSEC_OUTPUT_RAW_PRESSURE,
      BSEC_OUTPUT_RAW_HUMIDITY,
      BSEC_OUTPUT_RAW_GAS,
      BSEC_OUTPUT_IAQ,
      BSEC_OUTPUT_STATIC_IAQ,
      BSEC_OUTPUT_CO2_EQUIVALENT,
      BSEC_OUTPUT_BREATH_VOC_EQUIVALENT,
      BSEC_OUTPUT_SENSOR_HEAT_COMPENSATED_TEMPERATURE,
      BSEC_OUTPUT_SENSOR_HEAT_COMPENSATED_HUMIDITY,
  };

  iaqSensor.updateSubscription(sensorList, 10, BSEC_SAMPLE_RATE_LP);
  checkIaqSensorStatus();
}

void setup()
{
  Serial.begin(115200);
  delay(10);
  Serial.println();
  setupPms();
  setupBme();
  setupWifi();
  setupOTA();
  setupWebserver();
  setupMQTT();

  sensorData = new SimpleMap<String, double>([](String &a, String &b) -> int
                                             {
                                               if (a == b)
                                                 return 0;
                                               if (a > b)
                                                 return 1;
                                               /*if (a < b)*/ return -1;
                                             });
}

void readPms()
{
  pms.read();
  if (pms)
  { // successfull read
    sensorData->put("pm1", pms.pm01);
    sensorData->put("pm2p5", pms.pm25);
    sensorData->put("pm10", pms.pm10);
    sensorData->put("nc0p3", pms.n0p3);
    sensorData->put("nc0p5", pms.n0p5);
    sensorData->put("nc1", pms.n1p0);
    sensorData->put("nc2p5", pms.n2p5);
    sensorData->put("nc5", pms.n5p0);
    sensorData->put("nc10", pms.n10p0);
  }
  else
  {
    switch (pms.status)
    {
    case pms.OK: // should never come here
      break;     // included to compile without warnings
    case pms.ERROR_TIMEOUT:
      Serial.println(F(PMS_ERROR_TIMEOUT));
      break;
    case pms.ERROR_MSG_UNKNOWN:
      Serial.println(F(PMS_ERROR_MSG_UNKNOWN));
      break;
    case pms.ERROR_MSG_HEADER:
      Serial.println(F(PMS_ERROR_MSG_HEADER));
      break;
    case pms.ERROR_MSG_BODY:
      Serial.println(F(PMS_ERROR_MSG_BODY));
      break;
    case pms.ERROR_MSG_START:
      Serial.println(F(PMS_ERROR_MSG_START));
      break;
    case pms.ERROR_MSG_LENGTH:
      Serial.println(F(PMS_ERROR_MSG_LENGTH));
      break;
    case pms.ERROR_MSG_CKSUM:
      Serial.println(F(PMS_ERROR_MSG_CKSUM));
      break;
    case pms.ERROR_PMS_TYPE:
      Serial.println(F(PMS_ERROR_PMS_TYPE));
      break;
    }
  }
}

void logPms()
{
  Serial.printf("PM <=  1µm: %5d µg/m³\r\n", (int)sensorData->get("pm1"));
  Serial.printf("PM <=2.5µm: %5d µg/m³\r\n", (int)sensorData->get("pm2p5"));
  Serial.printf("PM <= 10µm: %5d µg/m³\r\n", (int)sensorData->get("pm10"));
  Serial.printf("NC >=0.3µm: %5d #/100cm³\r\n", (int)sensorData->get("nc0p3"));
  Serial.printf("NC >=0.5µm: %5d #/100cm³\r\n", (int)sensorData->get("nc0p5"));
  Serial.printf("NC >=  1µm: %5d #/100cm³\r\n", (int)sensorData->get("nc1"));
  Serial.printf("NC >=2.5µm: %5d #/100cm³\r\n", (int)sensorData->get("nc2p5"));
  Serial.printf("NC >=  5µm: %5d #/100cm³\r\n", (int)sensorData->get("nc5"));
  Serial.printf("NC >= 10µm: %5d #/100cm³\r\n", (int)sensorData->get("nc10"));
}

void readBme()
{
  if (iaqSensor.run())
  { // If new data is available
    sensorData->put("temperature", iaqSensor.temperature);
    sensorData->put("humidity", iaqSensor.humidity);
    sensorData->put("pressure", iaqSensor.pressure / 100); // Pa -> hPa
    sensorData->put("iaq", iaqSensor.iaq);
    sensorData->put("iaq_acc", (int)iaqSensor.iaqAccuracy);
    sensorData->put("siaq", iaqSensor.staticIaq);
    sensorData->put("siaq_acc", (int)iaqSensor.staticIaqAccuracy);
    sensorData->put("eco2", iaqSensor.co2Equivalent);
    sensorData->put("eco2_acc", (int)iaqSensor.co2Accuracy);
    sensorData->put("bvoc", iaqSensor.breathVocEquivalent);
    sensorData->put("bvoc_acc", (int)iaqSensor.breathVocAccuracy);
    updateBmeState();
  }
  else
  {
    checkIaqSensorStatus();
  }
}

void logBme()
{
  Serial.printf("Temperature  : %f °C\r\n", sensorData->get("temperature"));
  Serial.printf("Humidity     : %f %%\r\n", sensorData->get("humidity"));
  Serial.printf("Pressure     : %f hPa\r\n", sensorData->get("pressure"));
  Serial.printf("IAQ          : %f\r\n", sensorData->get("iaq"));
  Serial.printf("IAQ Accuracy : %d\r\n", (int)sensorData->get("iaq_acc"));
  Serial.printf("Static IAQ   : %f\r\n", sensorData->get("siaq"));
  Serial.printf("Stat IAQ Acc : %d\r\n", (int)sensorData->get("siaq_acc"));
  Serial.printf("CO2 Equiv    : %f ppm\r\n", sensorData->get("eco2"));
  Serial.printf("CO2 Accuracy : %d\r\n", (int)sensorData->get("eco2_acc"));
  Serial.printf("bVOC Equiv   : %f ppm\r\n", sensorData->get("bvoc"));
  Serial.printf("bVOC Accuracy: %d\r\n", (int)sensorData->get("bvoc_acc"));
  //Serial.printf("Gas Percent  : %f %%\r\n", iaqSensor.gasPercentage);
  //Serial.printf("Gas Per Accur: %d\r\n", iaqSensor.gasPercentageAcccuracy);
  //Serial.printf("Raw Temp     : %f °C\r\n", iaqSensor.rawTemperature);
  //Serial.printf("Raw Rel Humid: %f %%\r\n", iaqSensor.rawHumidity);
  //Serial.printf("Gas Resist   : %f Ohm\r\n", iaqSensor.gasResistance);
}

void publishMetric(String metricValues)
{
  String payloadString = "air,node=" + String(HOSTNAME) + " " + metricValues;
  //logLine(payloadString);
  int len_payload = payloadString.length() + 1;
  char payload[len_payload];
  payloadString.toCharArray(payload, len_payload);
  mqttClient.publish(MQTT_TOPIC, payload);
}

void readSensorData()
{
  readPms();
  readBme();
}

void sendMqttSensorData()
{
  String metricValues;
  for (int i = 0; i < sensorData->size(); i++)
  {
    metricValues += sensorData->getKey(i) + "=" + sensorData->getData(i);
    if (i < sensorData->size() - 1)
    {
      metricValues += ",";
    }
  }
  publishMetric(metricValues);
}

void logSensorData()
{
  logLine("\n----- PMS5003 -----");
  logPms();
  logLine("\n----- BME680 -----");
  logBme();
  logLine("\n#####");
}

void loop()
{
  ArduinoOTA.handle();
  webSocket.loop();
  server.handleClient();
  MDNS.update();

  static unsigned long readTimer = millis();
  if (millis() - readTimer >= 1000)
  {
    readSensorData();
    sendWebsocketData(0);
    readTimer = millis();
  }

  static unsigned long sendTimer = millis();
  if (millis() - sendTimer >= 5000)
  {
    if ((WiFi.status() != WL_CONNECTED))
    {
      logLine("Wifi not connected, reconnecting...");
      WiFi.disconnect();
      WiFi.reconnect();
    }
    if (!mqttClient.connected())
    {
      logLine("MQTT not connected, reconnecting...");
      connectMQTT();
    }

    sendMqttSensorData();
    logSensorData();
    sendTimer = millis();
  }

  mqttClient.loop();
}
